# Copyright (C) 2012 W. Trevor King <wking@tremily.us>
#
# This file is part of pypiezo.
#
# pypiezo is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pypiezo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pypiezo.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for wiggling a the piezo near the surface.

This helps you detect interference between the laser bouncing off the
tip and the surface.  The wiggling continues until you stop it, which
gives you feedback while you reposition the laser to minimize the
interference.  One day we'll all have fancy AFMs with software control
over the laser alignment, but until then, we've got this module
helping you with your thumbscrews.
"""

import time as _time

import numpy as _numpy

try:
    import h5py as _h5py
    import h5config as _h5config
    from h5config.storage.hdf5 import HDF5_Storage as _HDF5_Storage
    from h5config.storage.hdf5 import h5_create_group as _h5_create_group
except ImportError, e:
    _h5py = None
    _h5py_import_error = e

try:
    import matplotlib as _matplotlib
    import matplotlib.pyplot as _matplotlib_pyplot
except (ImportError, RuntimeError), e:
    _matplotlib = None
    _matplotlib_import_error = e

from curses_check_for_keypress import CheckForKeypress as _CheckForKeypress

from . import LOG as _LOG
from . import base as _base
from . import package_config as _package_config


def _setup_config(piezo, config):
    if config['wavelength'] and config['amplitude']:
        log_string = \
            'use either laser_wavelength or amplitude, but not both'
        _LOG.warn(log_string)

    if None in (config['amplitude'], config['offset']):
        output_axis = piezo.axis_by_name(config['axis'])
        maxdata = output_axis.axis_channel.get_maxdata()
        midpoint = int(maxdata/2)
        if config['offset'] is None:
            offset = midpoint
            _LOG.debug(('generated offset for interference wiggle: {}'
                        ).format(config['offset']))
        if config['amplitude'] is None:
            if config['offset'] <= midpoint:
                max_amplitude = int(config['offset'])
            else:
                max_amplitude = int(maxdata-config['offset'])
            offset_meters = _base.convert_bits_to_meters(
                output_axis.config, config['offset'])
            if config['wavelength'] is None:
                config['amplitude'] = 0.5*max_amplitude
            else:
                bit_wavelength = _base.convert_meters_to_bits(
                    output_axis.config,
                    offset_meters + config['wavelength']
                    ) - config['offset']
                config['amplitude'] = 2*bit_wavelength
            _LOG.debug(('generated amplitude for interference wiggle: {}'
                        ).format(config['amplitude']))
            if config['amplitude'] > max_amplitude:
                raise ValueError(
                    'no room for a two wavelength wiggle ({} > {})'.format(
                        config['amplitude'], max_amplitude))

def _construct_output(piezo, config):
    n = config['samples']  # samples in a single oscillation
    out = (config['amplitude']
           * _numpy.sin(_numpy.arange(2*n)*2*_numpy.pi/n)
           + config['offset'])
    # 2*n for 2 periods, so you can judge precision
    out = out.reshape((len(out), 1)).astype(
        piezo.channel_dtype(config['axis'], direction='output'))
    return out

def _setup_datafile(filename, group, piezo, config, output):
    if not _h5py:
        raise _h5py_import_error
    output_axis = afm.piezo.axis_by_name(config['axis'])
    input_channel = afm.piezo.input_channel_by_name(config['input'])
    with _h5py.File(filename, 'w') as f:
        cwg = _h5_create_group(f, group)
        storage = _HDF5_Storage()
        for config_,key in [
            (config, 'config/wiggle'),
            (output.axis.config,
             'config/{}/axis'.format(config['axis'])),
            (input_channel.config,
             'config/{}/channel'.format(config['input']))]:
            if config_ is None:
                continue
            config_cwg = _h5_create_group(cwg, key)
            storage.save(config=config, group=config_cwg)
        cwg['wiggle/raw/{}'.format(config['axis'])] = output

def _update_datafile(filename, group, config, cycle, data):
    timestamp = ('{0}-{1:02d}-{2:02d}T{3:02d}-{4:02d}-{5:02d}'
                 ).format(*_time.localtime())
    with _h5py.File(filename, 'a') as f:
        wiggle_group = _h5_create_group(f, group)
        cwg = _h5_create_group(
            wiggle_group, 'wiggle/{}'.format(cycle))
        cwg['time'] = timestamp
        cwg['raw/{}'.format(config['input'])] = data

def _setup_plot(piezo, config, output):
    if not _matplotlib:
        raise _matplotlib_import_error
    _matplotlib.interactive(True)
    figure = _matplotlib_pyplot.figure()
    axes = figure.add_subplot(1, 1, 1)
    axes.hold(False)
    timestamp = _time.strftime('%H%M%S')
    axes.set_title('wiggle for interference %s' % timestamp)
    plot = axes.plot(output, output, 'b.-')
    figure.canvas.draw()
    figure.show()
    if not _matplotlib.is_interactive():
        _matplotlib_pyplot.show()
    return (figure, axes, plot)

def _update_plot(figure, axes, plot, cycle, data):
    plot[0].set_ydata(data[:,0])
    axes.set_ylim([data.min(), data.max()])
    _matplotlib_pyplot.draw()

def _run_wiggles(piezo, config, figure, axes, plot, output, filename=None,
                 group='/', keypress_test_mode=False):
    scan_frequency = config['frequency'] * config['samples']
    cycle = 0
    c = _CheckForKeypress(test_mode=keypress_test_mode)
    while c.input() == None:
        # input will need processing for multi-segment AFMs...
        data = piezo.ramp(
            output, scan_frequency, output_names=[config['axis']],
            input_names=[config['input']])
        _LOG.debug('completed a wiggle round')
        if filename:
            _update_datafile(
                filename=filename, group=group, config=config,
                cycle=cycle, data=data)
        if plot:
            _update_plot(
                figure=figure, axes=axes, plot=plot, cycle=cycle, data=data)
        cycle += 1

def wiggle_for_interference(
    piezo, config, plot=True, filename=None, group='/',
    keypress_test_mode=False):
    """Output a sine wave and measure interference.

    With a poorly focused or aligned laser, leaked laser light
    reflecting off the surface may interfere with the light
    reflected off the cantilever, causing distance-dependent
    interference with a period roughly half the laser's
    wavelength.  This method wiggles the cantilever near the
    surface and monitors the magnitude of deflection oscillation,
    allowing the operator to adjust the laser alignment in order
    to minimize the interference.

    Modern commercial AFMs with computer-aligned lasers must do
    something like this automatically.
    """
    if _package_config['matplotlib']:
        plot = True
    _setup_config(piezo=piezo, config=config)
    _LOG.debug('oscillate for interference wiggle ({})'.format(config))
    output = _construct_output(piezo=piezo, config=config)

    if filename:
        _setup_datafile(
            filename=filename, group=group, piezo=piezo, config=config,
            output=output)
    if plot:
        interactive = _matplotlib.is_interactive()
        figure,axes,plot_ = _setup_plot(
            piezo=piezo, config=config, output=output)
    else:
        figure = axes = plot_ = None
    _run_wiggles(
        piezo=piezo, config=config, figure=figure, axes=axes, plot=plot_,
        output=output, filename=filename, group=group,
        keypress_test_mode=keypress_test_mode)
    if plot:
        _matplotlib.interactive(interactive)
    piezo.last_output[config['axis']] = output[-1,0]
    _LOG.debug('interference wiggle complete')



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of pypiezo.
#
# pypiezo is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pypiezo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pypiezo.  If not, see <http://www.gnu.org/licenses/>.

"""Basic piezo control.

Several of the classes defined in this module are simple wrappers for
combining a `pycomedi` class instance (e.g. `Channel`) with the
appropriate config data (e.g. `InputChannelConfig`).  The idea is that
the `h5config`-based config data will make it easy for you to save
your hardware configuration to disk (so that you have a record of what
you did).  It should also make it easy to load your configuration from
the disk (so that you can do the same thing again).  Because this
`h5config` <-> `pycomedi` communication works both ways, you have two
options when you're initializing a class:

1) On your first run, it's probably easiest to setup your channels and
   such using the usual `pycomedi` interface
   (`device.find_subdevice_by_type`, etc.)  After you have setup your
   channels, you can initialize them with a stock config instance, and
   call the `setup_config` method to copy the channel configuration
   into the config file.  Now the config instance is ready to be saved
   to disk.
2) On later runs, you have the option of loading the `pycomedi`
   objects from your old configuration.  After loading the config data
   from disk, initialize your class by passing in a `devices` list,
   but without passing in the `Channel` instances.  The class will
   take care of setting up the channel instances internally,
   recreating your earlier setup.

For examples of how to apply either approach to a particular class,
see that class' docstring.
"""

import math as _math
from time import sleep as _sleep

import numpy as _numpy
from scipy.stats import linregress as _linregress

try:
    import matplotlib as _matplotlib
    import matplotlib.pyplot as _matplotlib_pyplot
    import time as _time  # for timestamping lines on plots
except (ImportError, RuntimeError), e:
    _matplotlib = None
    _matplotlib_import_error = e

from pycomedi.device import Device
from pycomedi.subdevice import StreamingSubdevice
from pycomedi.channel import AnalogChannel
from pycomedi.constant import AREF, TRIG_SRC, SDF, SUBDEVICE_TYPE, UNIT
from pycomedi.utility import inttrig_insn, Reader, Writer

from . import LOG as _LOG
from . import config as _config
from . import package_config as _package_config


def convert_bits_to_volts(config, data):
    """Convert bit-valued data to volts.

    >>> config = _config.ChannelConfig()
    >>> config['conversion-coefficients'] = [1, 2, 3]
    >>> config['conversion-origin'] = -1
    >>> convert_bits_to_volts(config, -1)
    1
    >>> convert_bits_to_volts(
    ...     config, _numpy.array([-1, 0, 1, 2], dtype=_numpy.float))
    array([  1.,   6.,  17.,  34.])
    """
    coefficients = config['conversion-coefficients']
    origin = config['conversion-origin']
    return _numpy.polyval(list(reversed(coefficients)), data-origin)

def convert_volts_to_bits(config, data):
    """Convert bit-valued data to volts.

    >>> config = _config.ChannelConfig()
    >>> config['inverse-conversion-coefficients'] = [1, 2, 3]
    >>> config['inverse-conversion-origin'] = -1
    >>> convert_volts_to_bits(config, -1)
    1
    >>> convert_volts_to_bits(
    ...     config, _numpy.array([-1, 0, 1, 2], dtype=_numpy.float))
    array([  1.,   6.,  17.,  34.])

    Note that the inverse coeffiecient and offset are difficult to
    derive from the forward coefficient and offset.  The current
    Comedilib conversion functions, `comedi_to_physical()` and
    `comedi_from_physical()` take `comedi_polynomial_t` conversion
    arguments.  `comedi_polynomial_t` is defined in `comedilib.h`_,
    and holds a polynomial of length
    `COMEDI_MAX_NUM_POLYNOMIAL_COEFFICIENTS`, currently set to 4.  The
    inverse of this cubic polynomial might be another polynomial, but
    it might also have a more complicated form.  A general inversion
    solution is considered too complicated, so when you're setting up
    your configuration, you should use Comedilib to save both the
    forward and inverse coefficients and offsets.

    .. _comedilib.h: http://www.comedi.org/git?p=comedi/comedilib.git;a=blob;f=include/comedilib.h;hb=HEAD
    """
    origin = config['inverse-conversion-origin']
    inverse_coefficients = config['inverse-conversion-coefficients']
    if len(inverse_coefficients) == 0:
        raise NotImplementedError('cubic polynomial inversion')
    return _numpy.polyval(list(reversed(inverse_coefficients)), data-origin)

def convert_volts_to_meters(config, data):
    """Convert volt-valued data to meters.

    >>> config = _config.AxisConfig()
    >>> config['gain'] = 20.0
    >>> config['sensitivity'] = 8e-9
    >>> convert_volts_to_meters(config, 1)
    1.6e-07
    >>> convert_volts_to_meters(
    ...     config, _numpy.array([1, 6, 17, 34], dtype=_numpy.float))
    ... # doctest: +ELLIPSIS
    array([  1.6...e-07,   9.6...e-07,   2.7...e-06,
             5.4...e-06])
    """
    return data * config['gain'] * config['sensitivity']

def convert_meters_to_volts(config, data):
    """Convert bit-valued data to volts.

    >>> config = _config.AxisConfig()
    >>> config['gain'] = 20.0
    >>> config['sensitivity'] = 8e-9
    >>> convert_meters_to_volts(config, 1.6e-7)
    1.0
    >>> convert_meters_to_volts(
    ...     config, _numpy.array([1.6e-7, 9.6e-7, 2.72e-6, 5.44e-6],
    ...                          dtype=_numpy.float))
    array([  1.,   6.,  17.,  34.])
    """
    return data / (config['gain'] * config['sensitivity'])

def convert_bits_to_meters(axis_config, data):
    """Convert bit-valued data to meters.

    >>> channel_config = _config.ChannelConfig()
    >>> channel_config['conversion-coefficients'] = [1, 2, 3]
    >>> channel_config['conversion-origin'] = -1
    >>> axis_config = _config.AxisConfig()
    >>> axis_config['gain'] = 20.0
    >>> axis_config['sensitivity'] = 8e-9
    >>> axis_config['channel'] = channel_config
    >>> convert_bits_to_meters(axis_config, 1)
    ... # doctest: +ELLIPSIS
    2.7...e-06
    >>> convert_bits_to_meters(
    ...     axis_config, _numpy.array([-1, 0, 1, 2], dtype=_numpy.float))
    ... # doctest: +ELLIPSIS
    array([  1.6...e-07,   9.6...e-07,   2.7...e-06,
             5.4...e-06])
    """
    data = convert_bits_to_volts(axis_config['channel'], data)
    return convert_volts_to_meters(axis_config, data)

def convert_meters_to_bits(axis_config, data):
    """Convert meter-valued data to volts.

    >>> channel_config = _config.ChannelConfig()
    >>> channel_config['inverse-conversion-coefficients'] = [1, 2, 3]
    >>> channel_config['inverse-conversion-origin'] = -1
    >>> axis_config = _config.AxisConfig()
    >>> axis_config['gain'] = 20.0
    >>> axis_config['sensitivity'] = 8e-9
    >>> axis_config['channel'] = channel_config
    >>> convert_meters_to_bits(axis_config, 1.6e-7)
    17.0
    >>> convert_meters_to_bits(
    ...     axis_config,
    ...     _numpy.array([1.6e-7, 9.6e-7, 2.72e-6, 5.44e-6],
    ...                  dtype=_numpy.float))
    array([   17.,   162.,  1009.,  3746.])
    """
    data = convert_meters_to_volts(axis_config, data)
    return convert_volts_to_bits(axis_config['channel'], data)

def get_axis_name(axis_config):
    """Return the name of an axis from the `AxisConfig`

    This is useful, for example, with
    `Config.select_config(get_attribute=get_axis_name)`.
    """
    channel_config = axis_config['channel']
    return channel_config['name']

def load_device(filename, devices):
    """Return an open device from `devices` which has a given `filename`.

    Sometimes a caller will already have the required `Device`, in
    which case we just pull that instance out of `devices`, check that
    it's open, and return it.  Other times, the caller may want us to
    open the device ourselves, so if we can't find an appropriate
    device in `devices`, we create a new one, append it to `devices`
    (so the caller can close it later), and return it.

    You will have to open the `Device` yourself, though, because the
    open device instance should not be held by a particular
    `PiezoAxis` instance.  If you don't want to open devices yourself,
    you can pass in a blank list of devices, and the initialization
    routine will append any necessary-but-missing devices to it.

    >>> from pycomedi.device import Device

    >>> devices = [Device('/dev/comedi0')]
    >>> device = load_device(filename='/dev/comedi0', devices=devices)
    >>> device.filename
    '/dev/comedi0'
    >>> device.file is not None
    True
    >>> device.close()

    >>> devices = []
    >>> device = load_device(filename='/dev/comedi0', devices=devices)
    >>> devices == [device]
    True
    >>> device.filename
    '/dev/comedi0'
    >>> device.file is not None
    True
    >>> device.close()

    We try and return helpful errors when things go wrong:

    >>> device = load_device(filename='/dev/comedi0', devices=None)
    Traceback (most recent call last):
      ...
    TypeError: 'NoneType' object is not iterable
    >>> device = load_device(filename='/dev/comedi0', devices=tuple())
    Traceback (most recent call last):
      ...
    ValueError: none of the available devices ([]) match /dev/comedi0, and we cannot append to ()
    >>> device = load_device(filename='/dev/comediX', devices=[])
    Traceback (most recent call last):
      ...
    PyComediError: comedi_open (/dev/comediX): No such file or directory (None)
    """
    try:
        matching_devices = [d for d in devices if d.filename == filename]
    except TypeError:
        _LOG.error('non-iterable devices? ({})'.format(devices))
        raise
    if matching_devices:
        device = matching_devices[0]
        if device.file is None:
            device.open()
    else:
        device = Device(filename)
        device.open()
        try:
            devices.append(device)  # pass new device back to caller
        except AttributeError:
            device.close()
            raise ValueError(
                ('none of the available devices ({}) match {}, and we '
                 'cannot append to {}').format(
                    [d.filename for d in devices], filename, devices))
    return device

def _load_channel_from_config(channel, devices, subdevice_type):
    c = channel.config  # reduce verbosity
    if not channel.channel:
        device = load_device(filename=c['device'], devices=devices)
        if c['subdevice'] < 0:
            subdevice = device.find_subdevice_by_type(
                subdevice_type, factory=StreamingSubdevice)
        else:
            subdevice = device.subdevice(
                index=c['subdevice'], factory=StreamingSubdevice)
        channel.channel = subdevice.channel(
            index=c['channel'], factory=AnalogChannel,
            aref=c['analog-reference'])
        channel.channel.range = channel.channel.get_range(index=c['range'])
    channel.name = c['name']

def _setup_channel_config(config, channel):
    """Initialize the `ChannelConfig` `config` using the
    `AnalogChannel` `channel`.
    """
    config['device'] = channel.subdevice.device.filename
    config['subdevice'] = channel.subdevice.index
    config['channel'] = channel.index
    config['maxdata'] = channel.get_maxdata()
    config['range'] = channel.range.value
    config['analog-reference'] = AREF.index_by_value(channel.aref.value)
    converter = channel.get_converter()
    config['conversion-origin'
           ] = converter.get_to_physical_expansion_origin()
    config['conversion-coefficients'
           ] = converter.get_to_physical_coefficients()
    config['inverse-conversion-origin'
           ] = converter.get_from_physical_expansion_origin()
    config['inverse-conversion-coefficients'
           ] = converter.get_from_physical_coefficients()


class PiezoAxis (object):
    """A one-dimensional piezoelectric actuator.

    If used, the montoring channel must (as of now) be on the same
    device as the controlling channel.

    >>> from pycomedi.device import Device
    >>> from pycomedi.subdevice import StreamingSubdevice
    >>> from pycomedi.channel import AnalogChannel
    >>> from pycomedi.constant import (AREF, SUBDEVICE_TYPE, UNIT)

    >>> d = Device('/dev/comedi0')
    >>> d.open()

    >>> s_in = d.find_subdevice_by_type(SUBDEVICE_TYPE.ai,
    ...     factory=StreamingSubdevice)
    >>> s_out = d.find_subdevice_by_type(SUBDEVICE_TYPE.ao,
    ...     factory=StreamingSubdevice)

    >>> axis_channel = s_out.channel(
    ...     0, factory=AnalogChannel, aref=AREF.ground)
    >>> monitor_channel = s_in.channel(
    ...     0, factory=AnalogChannel, aref=AREF.diff)
    >>> for chan in [axis_channel, monitor_channel]:
    ...     chan.range = chan.find_range(unit=UNIT.volt, min=-10, max=10)

    >>> config = _config.AxisConfig()
    >>> config.update({'gain':20, 'sensitivity':8e-9})
    >>> config['channel'] = _config.OutputChannelConfig()
    >>> config['monitor'] = _config.InputChannelConfig()
    >>> config['monitor']['device'] = '/dev/comediX'

    >>> p = PiezoAxis(config=config)
    ... # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
      ...
    NotImplementedError: piezo axis control and monitor on different devices
     (/dev/comedi0 and /dev/comediX)

    >>> config['monitor']['device'] = config['channel']['device']
    >>> p = PiezoAxis(config=config,
    ...     axis_channel=axis_channel, monitor_channel=monitor_channel)

    >>> p.setup_config()
    >>> print(config['channel'].dump())
    name: 
    device: /dev/comedi0
    subdevice: 1
    channel: 0
    maxdata: 65535
    range: 0
    analog-reference: ground
    conversion-coefficients: -10.0, 0.000305180437934
    conversion-origin: 0.0
    inverse-conversion-coefficients: 0.0, 3276.75
    inverse-conversion-origin: -10.0
    >>> print(config['monitor'].dump())
    name: 
    device: /dev/comedi0
    subdevice: 0
    channel: 0
    maxdata: 65535
    range: 0
    analog-reference: diff
    conversion-coefficients: -10.0, 0.000305180437934
    conversion-origin: 0.0
    inverse-conversion-coefficients: 0.0, 3276.75
    inverse-conversion-origin: -10.0

    >>> convert_bits_to_meters(p.config, 0)
    ... # doctest: +ELLIPSIS
    -1.6...e-06

    Opening from the config alone:

    >>> p = PiezoAxis(config=config)
    >>> p.load_from_config(devices=[d])
    >>> p.axis_channel  # doctest: +ELLIPSIS
    <pycomedi.channel.AnalogChannel object at 0x...>
    >>> p.monitor_channel  # doctest: +ELLIPSIS
    <pycomedi.channel.AnalogChannel object at 0x...>

    >>> d.close()
    """
    def __init__(self, config, axis_channel=None, monitor_channel=None):
        self.config = config
        self.axis_channel = axis_channel
        self.monitor_channel = monitor_channel

    def load_from_config(self, devices):
        c = self.config  # reduce verbosity
        if (c['monitor'] and
            c['channel']['device'] != c['monitor']['device']):
            raise NotImplementedError(
                ('piezo axis control and monitor on different devices '
                 '({} and {})').format(
                    c['channel']['device'], c['monitor']['device']))
        if not self.axis_channel:
            output = OutputChannel(config=c['channel'])
            output.load_from_config(devices=devices)
            self.axis_channel = output.channel
        if c['monitor'] and not self.monitor_channel:
            monitor = InputChannel(config=c['monitor'])
            monitor.load_from_config(devices=devices)
            self.monitor_channel = monitor.channel
        self.name = c['channel']['name']

    def setup_config(self):
        "Initialize the axis (and monitor) configs."
        _setup_channel_config(self.config['channel'], self.axis_channel)
        if self.monitor_channel:
            _setup_channel_config(
                self.config['monitor'], self.monitor_channel)
        if self.config['minimum'] is None:
            self.config['minimum'] = convert_bits_to_volts(
                self.config['channel'], 0)
        if self.config['maximum'] is None:
            self.config['maximum'] = convert_bits_to_volts(
                self.config['channel'], self.axis_channel.get_maxdata())



class OutputChannel(object):
    """An input channel monitoring some interesting parameter.

    >>> from pycomedi.device import Device
    >>> from pycomedi.subdevice import StreamingSubdevice
    >>> from pycomedi.channel import AnalogChannel
    >>> from pycomedi.constant import AREF, SUBDEVICE_TYPE, UNIT

    >>> d = Device('/dev/comedi0')
    >>> d.open()

    >>> s = d.find_subdevice_by_type(SUBDEVICE_TYPE.ao,
    ...     factory=StreamingSubdevice)

    >>> channel = s.channel(0, factory=AnalogChannel, aref=AREF.diff)
    >>> channel.range = channel.find_range(unit=UNIT.volt, min=-10, max=10)

    >>> channel_config = _config.OutputChannelConfig()

    >>> c = OutputChannel(config=channel_config, channel=channel)
    >>> c.setup_config()
    >>> print(channel_config.dump())
    name: 
    device: /dev/comedi0
    subdevice: 1
    channel: 0
    maxdata: 65535
    range: 0
    analog-reference: diff
    conversion-coefficients: -10.0, 0.000305180437934
    conversion-origin: 0.0
    inverse-conversion-coefficients: 0.0, 3276.75
    inverse-conversion-origin: -10.0

    >>> convert_volts_to_bits(c.config, -10)
    0.0

    Opening from the config alone:

    >>> c = OutputChannel(config=channel_config)
    >>> c.load_from_config(devices=[d])
    >>> c.channel  # doctest: +ELLIPSIS
    <pycomedi.channel.AnalogChannel object at 0x...>

    >>> d.close()
    """
    def __init__(self, config, channel=None):
        self.config = config
        self.channel = channel

    def load_from_config(self, devices):
        _load_channel_from_config(
            channel=self, devices=devices, subdevice_type=SUBDEVICE_TYPE.ao)

    def setup_config(self):
        _setup_channel_config(self.config, self.channel)


class InputChannel(object):
    """An input channel monitoring some interesting parameter.

    >>> from pycomedi.device import Device
    >>> from pycomedi.subdevice import StreamingSubdevice
    >>> from pycomedi.channel import AnalogChannel
    >>> from pycomedi.constant import AREF, SUBDEVICE_TYPE, UNIT

    >>> d = Device('/dev/comedi0')
    >>> d.open()

    >>> s = d.find_subdevice_by_type(SUBDEVICE_TYPE.ai,
    ...     factory=StreamingSubdevice)

    >>> channel = s.channel(0, factory=AnalogChannel, aref=AREF.diff)
    >>> channel.range = channel.find_range(unit=UNIT.volt, min=-10, max=10)

    >>> channel_config = _config.InputChannelConfig()

    >>> c = InputChannel(config=channel_config, channel=channel)
    >>> c.setup_config()
    >>> print(channel_config.dump())
    name: 
    device: /dev/comedi0
    subdevice: 0
    channel: 0
    maxdata: 65535
    range: 0
    analog-reference: diff
    conversion-coefficients: -10.0, 0.000305180437934
    conversion-origin: 0.0
    inverse-conversion-coefficients: 0.0, 3276.75
    inverse-conversion-origin: -10.0

    >>> convert_bits_to_volts(c.config, 0)
    -10.0

    Opening from the config alone:

    >>> c = InputChannel(config=channel_config)
    >>> c.load_from_config(devices=[d])
    >>> c.channel  # doctest: +ELLIPSIS
    <pycomedi.channel.AnalogChannel object at 0x...>

    >>> d.close()
    """
    def __init__(self, config, channel=None):
        self.config = config
        self.channel = channel

    def load_from_config(self, devices):
        _load_channel_from_config(
            channel=self, devices=devices, subdevice_type=SUBDEVICE_TYPE.ai)

    def setup_config(self):
        _setup_channel_config(self.config, self.channel)


class Piezo (object):
    """A piezo actuator-controlled experiment.

    >>> from pprint import pprint
    >>> from pycomedi.device import Device
    >>> from pycomedi.subdevice import StreamingSubdevice
    >>> from pycomedi.channel import AnalogChannel
    >>> from pycomedi.constant import AREF, SUBDEVICE_TYPE, UNIT

    >>> d = Device('/dev/comedi0')
    >>> d.open()

    >>> axis_config = _config.AxisConfig()
    >>> axis_config['gain'] = 20.0
    >>> axis_config['sensitivity'] = 8e-9
    >>> axis_config['channel'] = _config.OutputChannelConfig()
    >>> axis_config['channel']['analog-reference'] = AREF.ground
    >>> axis_config['channel']['name'] = 'z'
    >>> axis_config['monitor'] = _config.InputChannelConfig()
    >>> axis_config['monitor']['analog-reference'] = AREF.diff
    >>> a = PiezoAxis(config=axis_config)
    >>> a.load_from_config(devices=[d])
    >>> a.setup_config()

    >>> input_config = _config.InputChannelConfig()
    >>> input_config['analog-reference'] = AREF.diff
    >>> input_config['name'] = 'some-input'
    >>> c = InputChannel(config=input_config)
    >>> c.load_from_config(devices=[d])
    >>> c.setup_config()

    >>> config = _config.PiezoConfig()
    >>> config['name'] = 'Charlie'

    >>> p = Piezo(config=config, axes=[a], inputs=[c])
    >>> p.setup_config()
    >>> inputs = p.read_inputs()
    >>> pprint(inputs)  # doctest: +SKIP
    {'some-input': 34494L, 'z-monitor': 32669L}

    >>> pos = convert_volts_to_bits(p.config['axes'][0]['channel'], 0)
    >>> pos
    32767.5
    >>> p.jump('z', pos)
    >>> p.last_output == {'z': int(pos)}
    True

    :meth:`ramp` raises an error if passed an invalid `data` `dtype`.

    >>> output_data = _numpy.arange(0, int(pos), step=int(pos/10),
    ...     dtype=_numpy.float)
    >>> output_data = output_data.reshape((len(output_data), 1))
    >>> input_data = p.ramp(data=output_data, frequency=10,
    ...     output_names=['z'], input_names=['z-monitor', 'some-input'])
    Traceback (most recent call last):
      ...
    ValueError: output dtype float64 does not match expected <type 'numpy.uint16'>
    >>> output_data = _numpy.arange(0, int(pos), step=int(pos/10),
    ...     dtype=p.channel_dtype('z', direction='output'))
    >>> output_data = output_data.reshape((len(output_data), 1))
    >>> input_data = p.ramp(data=output_data, frequency=10,
    ...     output_names=['z'], input_names=['z-monitor', 'some-input'])
    >>> input_data  # doctest: +SKIP
    array([[    0, 25219],
           [ 3101, 23553],
           [ 6384, 22341],
           [ 9664, 21465],
           [12949, 20896],
           [16232, 20614],
           [19516, 20588],
           [22799, 20801],
           [26081, 21233],
           [29366, 21870],
           [32646, 22686]], dtype=uint16)

    >>> p.last_output == {'z': output_data[-1]}
    True

    >>> data = p.named_ramp(data=output_data, frequency=10,
    ...     output_names=['z'], input_names=['z-monitor', 'some-input'])
    >>> pprint(data)  # doctest: +ELLIPSIS, +SKIP
    {'some-input': array([21666, 20566, ..., 22395], dtype=uint16),
     'z': array([    0,  3276,  ..., 32760], dtype=uint16),
     'z-monitor': array([ 3102,  6384,  ..., 32647], dtype=uint16)}

    Opening from the config alone:

    >>> p = Piezo(config=config)
    >>> p.load_from_config(devices=[d])
    >>> for axis in p.axes:
    ...     print(axis.axis_channel)
    ...     print(axis.monitor_channel)
    ... # doctest: +ELLIPSIS
    <pycomedi.channel.AnalogChannel object at 0x...>
    <pycomedi.channel.AnalogChannel object at 0x...>
    >>> for input in p.inputs:
    ...     print(input.channel)
    ... # doctest: +ELLIPSIS
    <pycomedi.channel.AnalogChannel object at 0x...>

    >>> d.close()
    """
    def __init__(self, config, axes=None, inputs=None):
        self.config=config
        self.axes = axes
        self.inputs = inputs
        self.last_output = {}

    def load_from_config(self, devices):
        if not self.axes:
            self.axes = []
            for config in self.config['axes']:
                axis = PiezoAxis(config=config)
                axis.load_from_config(devices=devices)
                self.axes.append(axis)
            self.last_output.clear()
        if not self.inputs:
            self.inputs = []
            for config in self.config['inputs']:
                input = InputChannel(config=config)
                input.load_from_config(devices=devices)
                self.inputs.append(input)
        self.name = self.config['name']

    def setup_config(self):
        "Initialize the axis and input configs."
        for x in self.axes + self.inputs:
            x.setup_config()
        self.config['axes'] = [x.config for x in self.axes]
        self.config['inputs'] = [x.config for x in self.inputs]

    def axis_by_name(self, name):
        "Get an axis by its name."
        for axis in self.axes:
            if axis.name == name:
                return axis
        raise ValueError(name)

    def input_channel_by_name(self, name):
        "Get an input channel by its name."
        for input_channel in self.inputs:
            if input_channel.name == name:
                return input_channel
        raise ValueError(name)

    def channels(self, direction=None):
        """Iterate through all `(name, channel)` tuples.

        ===========  ===================
        `direction`  Returned channels
        ===========  ===================
        'input'      all input channels
        'output'     all output channels
        None         all channels
        ===========  ===================
        """
        if direction not in ('input', 'output', None):
            raise ValueError(direction)
        for a in self.axes:
            if direction != 'input':
                yield (a.name, a.axis_channel)
            if a.monitor_channel and direction != 'output':
                yield ('%s-monitor' % a.name, a.monitor_channel)
        if direction != 'output':
            for c in self.inputs:
                yield (c.name, c.channel)

    def channel_by_name(self, name, direction=None):
        """Get a channel by its name.

        Setting `direction` (see :meth:`channels`) may allow a more
        efficient search.
        """
        for n,channel in self.channels(direction=direction):
            if n == name:
                return channel
        raise ValueError(name)

    def channel_dtype(self, channel_name, direction=None):
        """Get a channel's data type by name.

        Setting `direction` (see :meth:`channels`) may allow a more
        efficient search.
        """
        channel = self.channel_by_name(name=channel_name, direction=direction)
        return channel.subdevice.get_dtype()

    def read_inputs(self):
        "Read all inputs and return a `name`->`value` dictionary."
        # There is no multi-channel read instruction, so preform reads
        # sequentially.
        ret = dict([(n, c.data_read())
                    for n,c in self.channels(direction='input')])
        _LOG.debug('current position: %s' % ret)
        return ret

    def jump(self, axis_name, position, steps=1, sleep=None):
        "Move the output named `axis_name` to `position`."
        _LOG.debug('jump {} to {} in {} steps'.format(
                axis_name, position, steps))
        if steps > 1:
            try:
                orig_pos = self.last_output[axis_name]
            except KeyError, e:
                _LOG.warn(
                    ("cannot make a soft jump to {} because we don't have a "
                     'last-output position for {}').format(
                        position, axis_name))
                return self.jump(axis_name=axis_name, position=position)
            else:
                for i,pos in enumerate(_numpy.linspace(
                        orig_pos, position, steps+1)[1:]):
                    _LOG.debug('jump {} to {} ({} of {} steps)'.format(
                            axis_name, pos, i, steps))
                    self.jump(axis_name=axis_name, position=pos)
                    if sleep:
                        _sleep(sleep)
                return
        position = int(position)
        channel = self.channel_by_name(name=axis_name)
        channel.data_write(position)
        self.last_output[axis_name] = position

    def ramp(self, data, frequency, output_names, input_names=()):
        """Synchronized IO ramp writing `data` and reading `in_names`.

        Parameters
        ----------
        data : numpy array-like
            Row for each cycle, column for each output channel.
        frequency : float
            Target cycle frequency in Hz.
        output_names : list of strings
            Names of output channels in the same order as the columns
            of `data`.
        input_names : list of strings
            Names of input channels to monitor in the same order as
            the columns of the returned array.
        """
        if len(data.shape) != 2:
            raise ValueError(
                'ramp data must be two dimensional, not %d' % len(data.shape))
        if data.shape[1] != len(output_names):
            raise ValueError(
                'ramp data should have on column for each input, '
                'but has %d columns for %d inputs'
                % (data.shape[1], len(output_names)))
        n_samps = data.shape[0]
        log_string = 'ramp %d samples at %g Hz.  out: %s, in: %s' % (
            n_samps, frequency, output_names, input_names)
        _LOG.debug(log_string)  # _LOG on one line for easy commenting-out
        # TODO: check range?
        output_channels = [self.channel_by_name(name=n, direction='output')
                           for n in output_names]
        inputs = [self.channel_by_name(name=n, direction='input')
                          for n in input_names]

        ao_subdevice = output_channels[0].subdevice
        ai_subdevice = inputs[0].subdevice
        device = ao_subdevice.device

        output_dtype = ao_subdevice.get_dtype()
        if data.dtype != output_dtype:
            raise ValueError('output dtype %s does not match expected %s'
                             % (data.dtype, output_dtype))
        input_data = _numpy.ndarray(
            (n_samps, len(inputs)), dtype=ai_subdevice.get_dtype())

        _LOG.debug('setup ramp commands')
        scan_period_ns = int(1e9 / frequency)
        ai_cmd = ai_subdevice.get_cmd_generic_timed(
            len(inputs), scan_period_ns)
        ao_cmd = ao_subdevice.get_cmd_generic_timed(
            len(output_channels), scan_period_ns)

        ai_cmd.start_src = TRIG_SRC.int
        ai_cmd.start_arg = 0
        ai_cmd.stop_src = TRIG_SRC.count
        ai_cmd.stop_arg = n_samps
        ai_cmd.chanlist = inputs
        #ao_cmd.start_src = TRIG_SRC.ext
        #ao_cmd.start_arg = 18  # NI card AI_START1 internal AI start signal
        ao_cmd.start_src = TRIG_SRC.int
        ao_cmd.start_arg = 0
        ao_cmd.stop_src = TRIG_SRC.count
        ao_cmd.stop_arg = n_samps-1
        ao_cmd.chanlist = output_channels

        ai_subdevice.cmd = ai_cmd
        ao_subdevice.cmd = ao_cmd
        for i in range(3):
            rc = ai_subdevice.command_test()
            if rc is None: break
            _LOG.debug('analog input test %d: %s' % (i, rc))
        for i in range(3):
            rc = ao_subdevice.command_test()
            if rc is None: break
            _LOG.debug('analog output test %d: %s' % (i, rc))

        _LOG.debug('lock subdevices for ramp')
        ao_subdevice.lock()
        try:
            ai_subdevice.lock()
            try:
                _LOG.debug('load ramp commands')
                ao_subdevice.command()
                ai_subdevice.command()

                writer = Writer(ao_subdevice, data)
                writer.start()
                reader = Reader(ai_subdevice, input_data)
                reader.start()

                _LOG.debug('arm analog output')
                device.do_insn(inttrig_insn(ao_subdevice))
                _LOG.debug('trigger ramp (via analog input)')
                device.do_insn(inttrig_insn(ai_subdevice))
                _LOG.debug('ramp running')

                writer.join()
                reader.join()
                _LOG.debug('ramp complete')
            finally:
                #_LOG.debug('AI flags: %s' % ai_subdevice.get_flags())
                ai_subdevice.cancel()
                ai_subdevice.unlock()
        finally:
            # release busy flag, which seems to not be cleared
            # automatically.  See
            #   http://groups.google.com/group/comedi_list/browse_thread/thread/4c7040989197abad/
            #_LOG.debug('AO flags: %s' % ao_subdevice.get_flags())
            ao_subdevice.cancel()
            ao_subdevice.unlock()
            _LOG.debug('unlocked subdevices after ramp')

        for i,name in enumerate(output_names):
            self.last_output[name] = data[-1,i]

        if _package_config['matplotlib']:
            if not _matplotlib:
                raise _matplotlib_import_error
            figure = _matplotlib_pyplot.figure()
            axes = figure.add_subplot(1, 1, 1)
            axes.hold(True)
            timestamp = _time.strftime('%H%M%S')
            axes.set_title('piezo ramp %s' % timestamp)
            for d,names in [(data, output_names),
                            (input_data, input_names)]:
                for i,name in enumerate(names):
                    axes.plot(d[:,i], label=name)
            figure.canvas.draw()
            figure.show()
            if not _matplotlib.is_interactive():
                _matplotlib_pyplot.show()
        return input_data

    def named_ramp(self, data, frequency, output_names, input_names=()):
        input_data = self.ramp(
            data=data, frequency=frequency, output_names=output_names,
            input_names=input_names)
        ret = {}
        for i,name in enumerate(output_names):
            ret[name] = data[:,i]
        for i,name in enumerate(input_names):
            ret[name] = input_data[:,i]
        return ret

    def zero(self, axis_names=None, **kwargs):
        zeros = []
        if axis_names is None:
            axis_names = [axis.name for axis in self.axes]
        for axis_name in axis_names:
            axis = self.axis_by_name(axis_name)
            config = self.config.select_config(
                'axes', axis_name, get_attribute=get_axis_name)['channel']
            zero = convert_volts_to_bits(config, 0)
            zeros.append(zero)
            self.jump(axis_name, zero)
        return zeros



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of pypiezo.
#
# pypiezo is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pypiezo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pypiezo.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities detecting the position of the sample surface.

TODO: doctests
"""

SLEEP_DURING_SURF_POS_AQUISITION = False # doesn't help

#from time import sleep as _sleep
import numpy as _numpy
from scipy.optimize import leastsq as _leastsq

try:
    import matplotlib as _matplotlib
    import matplotlib.pyplot as _matplotlib_pyplot
    import time as _time  # for timestamping lines on plots
except (ImportError, RuntimeError), e:
    _matplotlib = None
    _matplotlib_import_error = e

from . import LOG as _LOG
from . import package_config as _package_config
from . import base as _base


class SurfaceError (Exception):
    pass


class PoorFit (SurfaceError):
    pass


class PoorGuess (PoorFit):
    pass


class FlatFit (PoorFit):
    "Raised for slope in the non-contact region, or no slope in contact region"
    def __init__(self, left_slope, right_slope):
        self.left_slope = left_slope
        self.right_slope = right_slope
        msg = 'slopes not sufficiently different: %g and %g' % (
            left_slope, right_slope)
        super(FlatFit, self).__init__(msg)


class EdgeKink (PoorFit):
    def __init__(self, kink, edge, window):
        self.kink = kink
        self.edge = edge
        self.window = window
        msg = 'no kink (kink %d not within %d of %d)' % (kink, window, edge)
        super(EdgeKink, self).__init__(self, msg)


def _linspace(*args, **kwargs):
    dtype = kwargs.pop('dtype')
    out = _numpy.linspace(*args, **kwargs)
    return out.reshape((len(out), 1)).astype(dtype)

def _get_min_max_positions(piezo, axis_name, min_position=None,
                           max_position=None):
    output_axis = piezo.axis_by_name(axis_name)
    if min_position is None:
        min_position = _base.convert_volts_to_bits(
            output_axis.config['channel'],
            output_axis.config['minimum'])
    if max_position is None:
        max_position = _base.convert_volts_to_bits(
            output_axis.config['channel'],
            output_axis.config['maximum'])
    return (min_position, max_position)

def get_surface_position_data(piezo, axis_name, max_deflection, steps=2000,
                              frequency=10e3, min_position=None,
                              max_position=None):
    "Measure the distance to the surface"
    _LOG.info('get surface position')
    orig_position = piezo.last_output[axis_name]
    # fully retract the piezo
    min_position,max_position = _get_min_max_positions(
        piezo, axis_name, min_position=min_position, max_position=max_position)
    _LOG.info('retract the piezo to %d' % min_position)
    dtype = piezo.channel_dtype(axis_name, direction='output')
    out = _linspace(orig_position, min_position, steps, dtype=dtype)
    out = out.reshape((len(out), 1)).astype(
        piezo.channel_dtype(axis_name, direction='output'))
    ramp_kwargs = {
        'frequency': frequency,
        'output_names': [axis_name],
        'input_names': ['deflection'],
        }
    ret1 = piezo.named_ramp(data=out, **ramp_kwargs)
    # locate high deflection position
    _LOG.info('approach until there is dangerous deflection (> %d)'
               % max_deflection)
    if SLEEP_DURING_SURF_POS_AQUISITION == True:
        _sleep(.2) # sleeping briefly seems to reduce bounce?
    mtpod = piezo.move_to_pos_or_def(
        axis_name=axis_name, position=max_position, deflection=max_deflection,
        step=(max_position-min_position)/steps, return_data=True)
    high_contact_pos = piezo.last_output[axis_name]
    # fully retract the piezo again
    _LOG.info('retract the piezo to %d again' % min_position)
    if SLEEP_DURING_SURF_POS_AQUISITION == True:
        _sleep(.2)
    out = _linspace(high_contact_pos, min_position, steps, dtype=dtype)
    ret2 = piezo.named_ramp(data=out, **ramp_kwargs)
    # scan to the high contact position
    _LOG.info('ramp in to the deflected position %d' % high_contact_pos)
    if SLEEP_DURING_SURF_POS_AQUISITION == True:
        _sleep(.2)
    out = _linspace(min_position, high_contact_pos, steps, dtype=dtype)
    data = piezo.named_ramp(data=out, **ramp_kwargs)
    if SLEEP_DURING_SURF_POS_AQUISITION == True:
        _sleep(.2)
    # return to the original position
    _LOG.info('return to the original position %d' % orig_position)
    out = _linspace(high_contact_pos, orig_position, steps, dtype=dtype)
    ret3 = piezo.named_ramp(data=out, **ramp_kwargs)
    return {'ret1':ret1, 'mtpod':mtpod, 'ret2':ret2,
            'approach':data, 'ret3':ret3}

def bilinear(x, params):
    """bilinear fit for surface bumps.  Model has two linear regimes
    which meet at x=kink_position and have independent slopes.

    `x` should be a `numpy.ndarray`.
    """
    left_offset,left_slope,kink_position,right_slope = params
    left_mask = x < kink_position
    right_mask = x >= kink_position # = not left_mask
    left_y = left_offset + left_slope*x
    right_y = (left_offset + left_slope*kink_position
               + right_slope*(x-kink_position))
    return left_mask * left_y + right_mask * right_y

def analyze_surface_position_data(
    ddict, min_slope_ratio=10.0, kink_window=None,
    return_all_parameters=False):
    """

    min_slope_ratio : float
        Minimum ratio between the non-contact "left" slope and the
        contact "right" slope.
    kink_window : int (in bits) or None
        Raise `EdgeKink` if the kink is within `kink_window` of the
        minimum or maximum `z` position during the approach.  If
        `None`, a default value of 2% of the approach range is used.
    """
    # uses ddict["approach"] for analysis
    # the others are just along to be plotted
    _LOG.info('analyze surface position data')

    data = ddict['approach']
    # analyze data, using bilinear model
    # y = p0 + p1 x                for x <= p2
    #   = p0 + p1 p2 + p3 (x-p2)   for x >= p2
    dump_before_index = 0 # 25 # HACK!!
    # Generate a reasonable guess...
    start_pos = data['z'].min()
    final_pos = data['z'].max()
    if final_pos == start_pos:
        raise SurfaceError('cannot fit a single approach step (too close?)')
    start_def = data['deflection'].min()
    final_def = data['deflection'].max()
    # start_def and start_pos are probably for 2 different points
    _LOG.info('min deflection {}, max deflection {}'.format(
            start_def, final_def))
    _LOG.info('min position {}, max position {}'.format(start_pos, final_pos))

    left_offset   = start_def
    left_slope    = 0
    kink_position = (final_pos+start_pos)/2.0
    right_slope   = 2.0*(final_def-start_def)/(final_pos-start_pos)
    pstart = [left_offset, left_slope, kink_position, right_slope]
    _LOG.info('guessed params: %s' % pstart)

    offset_scale = (final_pos - start_pos)/100
    left_slope_scale = right_slope/10
    kink_scale = (final_pos-start_pos)/100
    right_slope_scale = right_slope
    scale = [offset_scale, left_slope_scale, kink_scale, right_slope_scale]
    _LOG.info('guessed scale: %s' % scale)

    def residual(p, y, x):
        Y = bilinear(x, p)
        return Y-y
    params,cov,info,mesg,ier = _leastsq(
        residual, pstart,
        args=(data["deflection"][dump_before_index:],
              data["z"][dump_before_index:]),
        full_output=True, maxfev=10000)
    _LOG.info('best fit parameters: %s' % (params,))

    if _package_config['matplotlib']:
        if not _matplotlib:
            raise _matplotlib_import_error
        figure = _matplotlib_pyplot.figure()
        axes = figure.add_subplot(1, 1, 1)
        axes.hold(True)
        timestamp = _time.strftime('%H%M%S')
        axes.set_title('surf_pos %5g %5g %5g %5g' % tuple(params))
        def plot_dict(d, label):
            _pylab.plot(d["z"], d["deflection"], '.',label=label)
        for n,name in [('ret1', 'first retract'),
                       ('mtpod', 'single step in'),
                       ('ret2', 'second retract'),
                       ('approach', 'main approach'),
                       ('ret3', 'return to start')]:
            axes.plot(ddict[n]['z'], ddict[n]['deflection'], label=name)
        def fit_fn(x, params):
            if x <= params[2]:
                return params[0] + params[1]*x
            else:
                return (params[0] + params[1]*params[2]
                        + params[3]*(x-params[2]))
        axes.plot([start_pos, params[2], final_pos],
                  [fit_fn(start_pos, params), fit_fn(params[2], params),
                   fit_fn(final_pos, params)], '-',label='fit')
        #_pylab.legend(loc='best')
        figure.canvas.draw()
        figure.show()
        if not _matplotlib.is_interactive():
            _matplotlib_pyplot.show()

    # check that the fit is reasonable
    # params[1] is slope in non-contact region
    # params[2] is kink position
    # params[3] is slope in contact region
    if kink_window is None:
        kink_window = int(0.02*(final_pos-start_pos))

    if abs(params[1]*min_slope_ratio) > abs(params[3]):
        raise FlatFit(left_slope=params[1], right_slope=params[3])
    if params[2] < start_pos+kink_window:
        raise EdgeKink(kink=params[2], edge=start_pos, window=kink_window)
    if params[2] > final_pos-kink_window:
        raise EdgeKink(kink=params[2], edge=final_pos, window=kink_window)
    _LOG.info('surface position %s' % params[2])
    if return_all_parameters:
        return params
    return params[2]

def get_surface_position(piezo, axis_name, max_deflection, **kwargs):
    ddict = get_surface_position_data(piezo, axis_name, max_deflection)
    return analyze_surface_position_data(ddict, **kwargs)



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of pypiezo.
#
# pypiezo is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pypiezo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pypiezo.  If not, see <http://www.gnu.org/licenses/>.

"Control of a piezo-based atomic force microscope."

import time as _time

import numpy as _numpy

try:
    import matplotlib as _matplotlib
    import matplotlib.pyplot as _matplotlib_pyplot
except (ImportError, RuntimeError), e:
    _matplotlib = None
    _matplotlib_import_error = e

from . import LOG as _LOG
from . import base as _base
from . import package_config as _package_config
from . import surface as _surface
from . import wiggle as _wiggle


class AFMPiezo (_base.Piezo):
    """A piezo-controlled atomic force microscope.

    This particular class expects a single input channel for measuring
    deflection.  Other subclasses provide support for multi-segment
    deflection measurements.

    >>> from pprint import pprint
    >>> from pycomedi.constant import AREF
    >>> from . import config
    >>> from . import surface

    >>> devices = []

    >>> piezo_config = config.PiezoConfig()
    >>> piezo_config['name'] = 'Molly'
    >>> piezo_config['axes'] = [config.AxisConfig()]
    >>> piezo_config['axes'][0]['channel'] = config.OutputChannelConfig()
    >>> piezo_config['axes'][0]['channel']['analog-reference'] = AREF.ground
    >>> piezo_config['axes'][0]['channel']['name'] = 'z'
    >>> piezo_config['inputs'] = [config.InputChannelConfig()]
    >>> piezo_config['inputs'][0]['analog-reference'] = AREF.diff
    >>> piezo_config['inputs'][0]['name'] = 'deflection'

    We set the minimum voltage for the `z` axis to -9 (a volt above
    the minimum possible voltage) to help with testing
    `.get_surface_position`.  Without this minimum voltage, small
    calibration errors could lead to a railed -10 V input for the
    first few surface approaching steps, which could lead to an
    `EdgeKink` error instead of a `FlatFit` error.

    >>> piezo_config['axes'][0].update(
    ...     {'gain':20, 'sensitivity':8e-9, 'minimum':-9})

    >>> p = AFMPiezo(config=piezo_config)
    >>> p.load_from_config(devices=devices)
    >>> p.setup_config()

    >>> deflection = p.read_deflection()
    >>> deflection  # doctest: +SKIP
    34494L
    >>> p.deflection_dtype()
    <type 'numpy.uint16'>

    We need to know where we are before we can move somewhere
    smoothly.

    >>> zeros = p.zero(axis_names=['z'])
    >>> pos = zeros[0]

    Usually `.move_to_pos_or_def` is used to approach the surface, but
    for testing we assume the z output channel is connected directly
    into the deflection input channel.

    >>> target_pos = _base.convert_volts_to_bits(
    ...     p.config.select_config('axes', 'z',
    ...     get_attribute=_base.get_axis_name)['channel'], 2)
    >>> step = int((target_pos - pos)/5)
    >>> target_def = _base.convert_volts_to_bits(
    ...     p.config.select_config('inputs', 'deflection'), 3)
    >>> data = p.move_to_pos_or_def('z', target_pos, target_def, step=step,
    ...     return_data=True)
    >>> p.last_output == {'z': int(target_pos)}
    True
    >>> pprint(data)  # doctest: +SKIP
    {'deflection':
       array([32655, 33967, 35280, 36593, 37905, 39218, 39222], dtype=uint16),
     'z':
       array([32767, 34077, 35387, 36697, 38007, 39317, 39321], dtype=uint16)}

    That was a working position-limited approach.  Now move back to
    the center and try a deflection-limited approach.

    >>> p.jump('z', pos)
    >>> target_def = _base.convert_volts_to_bits(
    ...     p.config.select_config('inputs', 'deflection'), 1)
    >>> data = p.move_to_pos_or_def('z', target_pos, target_def, step=step,
    ...     return_data=True)
    >>> print (p.last_output['z'] < int(target_pos))
    True
    >>> pprint(data)  # doctest: +SKIP
    {'deflection': array([32655, 33968, 35281, 36593], dtype=uint16),
     'z': array([32767, 34077, 35387, 36697], dtype=uint16)}

    >>> wiggle_config = config.WiggleConfig()
    >>> wiggle_config['offset'] = p.last_output['z']
    >>> wiggle_config['wavelength'] = 650e-9
    >>> p.wiggle_for_interference(config=wiggle_config,
    ...     keypress_test_mode=True)
    Press any key to continue

    >>> try:
    ...     p.get_surface_position('z', max_deflection=target_def)
    ... except surface.FlatFit, e:
    ...     print 'got FlatFit'
    got FlatFit
    >>> print e  # doctest: +SKIP
    slopes not sufficiently different: 1.0021 and 1.0021
    >>> abs(e.right_slope-1) < 0.1
    True
    >>> abs(e.left_slope-1) < 0.1
    True

    >>> for device in devices:
    ...     device.close()
    """
    def _deflection_channel(self):
        return self.channel_by_name(name='deflection', direction='input')

    def read_deflection(self):
        """Return sensor deflection in bits.

        TODO: explain how bit <-> volt conversion will work for this
        "virtual" channel.
        """
        return self._deflection_channel().data_read()

    def deflection_dtype(self):
        "Return a Numpy dtype suitable for deflection bit values."
        return self._deflection_channel().subdevice.get_dtype()

    def move_to_pos_or_def(self, axis_name, position=None, deflection=None,
                           step=1, return_data=False, pre_move_steps=0,
                           frequency=None):
        """TODO

        pre_move_steps : int
            number of 'null' steps to take before moving (confirming a
            stable input deflection).
        frequency : float
            The target step frequency in hertz.  If `Null`, go as fast
            as possible.  Note that this is software timing, so it
            should not be relied upon for precise results.
        """
        if position is None and deflection is None:
            raise ValueError('must specify position, deflection, or both')

        if return_data or _package_config['matplotlib']:
            aquire_data = True
        else:
            aquire_data = False

        if position is None:
            # default to the extreme value in the step direction
            if step > 0:
                axis = self.axis_by_name(axis_name)
                position = axis.axis_channel.get_maxdata()
            else:
                position = 0
        elif deflection is None:
            # default to the extreme value
            channel = self._deflection_channel(self)
            deflection = channel.get_maxdata()
        position = int(position)  # round down to nearest integer

        if step == 0:
            raise ValueError('must have non-zero step size')
        elif step < 0 and position > self.last_output[axis_name]:
            step = -step
        elif step > 0 and position < self.last_output[axis_name]:
            step = -step

        log_string = (
            'move to position %d or deflection %g on axis %s in steps of %d'
            % (position, deflection, axis_name, step))
        _LOG.debug(log_string)
        current_deflection = self.read_deflection()
        log_string = 'current position %d and deflection %g' % (
            self.last_output[axis_name], current_deflection)
        _LOG.debug(log_string)

        if aquire_data:
            def_array=[current_deflection]
            pos_array=[self.last_output[axis_name]]
        for i in range(pre_move_steps):
            self.jump(axis_name, piezo.last_output[axis_name])
            current_deflection = self.read_deflection()
            if aquire_data:
                def_array.append(current_deflection)
                pos_array.append(self.last_output[axis_name])
        if frequency is not None:
            time_step = 1./frequency
            next_time = _time.time() + time_step
        # step in until we hit our target position or exceed our target deflection
        while (self.last_output[axis_name] != position and
               current_deflection < deflection):
            dist_to = position - self.last_output[axis_name]
            if abs(dist_to) < abs(step):
                jump_to = position
            else:
                jump_to = self.last_output[axis_name] + step
            self.jump(axis_name, jump_to)
            current_deflection = self.read_deflection()
            log_string = (
                ('current z piezo position {} (target {}), '
                 'current deflection {} (target {})').format(
                    self.last_output[axis_name], position,
                    current_deflection, deflection))
            _LOG.debug(log_string)
            if aquire_data:
                def_array.append(current_deflection)
                pos_array.append(self.last_output[axis_name])
            if frequency is not None:
                now = _time.time()
                if now < next_time:
                    _time.sleep(next_time - now)
                next_time += time_step

        log_string = (
            'move to position %d or deflection %g on axis %s complete'
            % (position, deflection, axis_name))
        _LOG.debug(log_string)
        log_string = 'current position %d and deflection %g' % (
            self.last_output[axis_name], current_deflection)
        _LOG.debug(log_string)
        if _package_config['matplotlib']:
            if not _matplotlib:
                raise _matplotlib_import_error
            figure = _matplotlib_pyplot.figure()
            axes = figure.add_subplot(1, 1, 1)
            axes.hold(True)
            timestamp = _time.strftime('%H%M%S')
            axes.set_title('step approach %s' % timestamp)
            axes.plot(pos_array, def_array, '.', label=timestamp)
            #_pylab.legend(loc='best')
            figure.canvas.draw()
            figure.show()
            if not _matplotlib.is_interactive():
                _matplotlib_pyplot.show()

        if return_data:
            data = {
                axis_name:_numpy.array(
                    pos_array, dtype=self.channel_dtype(
                        axis_name, direction='output')),
                'deflection':_numpy.array(
                    def_array, dtype=self.deflection_dtype()),
                }
            return data

    wiggle_for_interference = _wiggle.wiggle_for_interference
    get_surface_position = _surface.get_surface_position


#def ramp
#        if USE_ABCD_DEFLECTION :
#            for i in range(4) : # i is the photodiode element (0->A, 1->B, ...)
#                self.curIn[i] = out["Deflection segment"][i][-1]
#        else :
#            self.curIn[self.chan_info.def_ind] = out["deflection"][-1]


#class FourSegmentAFM (AFM):
#    def read_deflection(self):
#        "Return sensor deflection in bits."
#        A = int(self.curIn[self.chan_info.def_ind[0]])
#        B = int(self.curIn[self.chan_info.def_ind[1]])
#        C = int(self.curIn[self.chan_info.def_ind[2]])
#        D = int(self.curIn[self.chan_info.def_ind[3]])
#        df = float((A+B)-(C+D))/(A+B+C+D)
#        dfout = int(df * 2**15) + 2**15
#        if TEXT_VERBOSE :
#            print "Current deflection %d (%d, %d, %d, %d)" \
#                % (dfout, A, B, C, D)
#        return dfout


#def test_smoothness(zp, plotVerbose=True):
#    posA = 20000
#    posB = 50000
#    setpoint = zp.def_V2in(3)
#    steps = 200
#    outfreq = 1e5
#    outarray = linspace(posB, posA, 1000)
#    indata=[]
#    outdata=[]
#    curVals = zp.jumpToPos(posA)
#    zp.pCurVals(curVals)
#    _sleep(1) # let jitters die down
#    for i in range(10):
#        print "ramp %d to %d" % (zp.curPos(), posB)
#        curVals, data = moveToPosOrDef(zp, posB, setpoint, step=steps,
#                                       return_data = True)
#        indata.append(data)
#        out = zp.ramp(outarray, outfreq)
#        outdata.append(out)
#    if plotVerbose:
#        from pylab import figure, plot, title, legend, hold, subplot        
#    if PYLAB_VERBOSE or plotVerbose:
#        _import_pylab()
#        _pylab.figure(BASE_FIG_NUM+4)
#        for i in range(10):
#            _pylab.plot(indata[i]['z'],
#                        indata[i]['deflection'], '+--', label='in')
#            _pylab.plot(outdata[i]['z'],
#                        outdata[i]['deflection'], '.-', label='out')
#        _pylab.title('test smoothness (step in, ramp out)')
#        #_pylab.legend(loc='best')
#    
#def test():
#    import z_piezo
#    zp = z_piezo.z_piezo()
#    curVals = zp.moveToPosOrDef(zp.pos_nm2out(600), defl=zp.curDef()+6000, step=(zp.pos_nm2out(10)-zp.pos_nm2out(0)))
#    if TEXT_VERBOSE:
#        zp.pCurVals(curVals)
#    pos = zp.getSurfPos(maxDefl=zp.curDef()+6000)
#    if TEXT_VERBOSE:
#        print "Surface at %g nm", pos
#    print "success"
#    if PYLAB_VERBOSE and _final_flush_plot != None:
#        _final_flush_plot()




# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of pypiezo.
#
# pypiezo is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pypiezo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pypiezo.  If not, see <http://www.gnu.org/licenses/>.

from .config import PackageConfig as _PackageConfig


__version__ = '0.7'

package_config = _PackageConfig(package_name=__name__)
package_config.load_system()



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of pypiezo.
#
# pypiezo is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pypiezo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pypiezo.  If not, see <http://www.gnu.org/licenses/>.

"""Construct example piezo configurations to simplify testing

These methods make it easier to write tests in packages that depend on
pypiezo.
"""

from .base import get_axis_name as _get_axis_name
from .config import AxisConfig as _AxisConfig
from .config import InputChannelConfig as _InputChannelConfig
from .config import OutputChannelConfig as _OutputChannelConfig
from .config import PiezoConfig as _PiezoConfig


def get_piezo_config(storage=None):
    """Return a default PiezoConfig instance.

    >>> p = get_piezo_config()
    >>> print p.dump()  # doctest: +REPORT_UDIFF
    name: test piezo
    axes:
      0:
        gain: 1.0
        sensitivity: 1.0
        minimum: None
        maximum: None
        channel:
          name: x
          device: /dev/comedi0
          subdevice: -1
          channel: 0
          maxdata: 100
          range: 0
          analog-reference: ground
          conversion-coefficients: 0, 1
          conversion-origin: 0
          inverse-conversion-coefficients: 0, 1
          inverse-conversion-origin: 0
        monitor: 
      1:
        gain: 1.0
        sensitivity: 1.0
        minimum: None
        maximum: None
        channel:
          name: z
          device: /dev/comedi0
          subdevice: -1
          channel: 1
          maxdata: 100
          range: 0
          analog-reference: ground
          conversion-coefficients: 0, 1
          conversion-origin: 0
          inverse-conversion-coefficients: 0, 1
          inverse-conversion-origin: 0
        monitor: 
    inputs:
      0:
        name: deflection
        device: /dev/comedi0
        subdevice: -1
        channel: 0
        maxdata: 100
        range: 0
        analog-reference: ground
        conversion-coefficients: 0, 1
        conversion-origin: 0
        inverse-conversion-coefficients: 0, 1
        inverse-conversion-origin: 0
      1:
        name: temperature
        device: /dev/comedi0
        subdevice: -1
        channel: 1
        maxdata: 100
        range: 0
        analog-reference: ground
        conversion-coefficients: 0, 1
        conversion-origin: 0
        inverse-conversion-coefficients: 0, 1
        inverse-conversion-origin: 0
    >>> _get_axis_name(p['axes'][0])
    'x'
    >>> d = p.select_config('inputs', 'deflection')
    >>> d['name']
    'deflection'
    """
    piezo_config = _PiezoConfig(storage=storage)
    piezo_config['name'] = 'test piezo'
    piezo_config['axes'] = []
    for i,name in enumerate(['x', 'z']):
        axis = _AxisConfig()
        axis['channel'] = _OutputChannelConfig()
        axis['channel']['name'] = name
        axis['channel']['channel'] = i
        axis['channel']['maxdata'] = 100
        axis['channel']['conversion-coefficients'] = (0,1)
        axis['channel']['conversion-origin'] = 0
        axis['channel']['inverse-conversion-coefficients'] = (0,1)
        axis['channel']['inverse-conversion-origin'] = 0
        piezo_config['axes'].append(axis)
    piezo_config['inputs'] = []
    for i,name in enumerate(['deflection', 'temperature']):
        channel = _InputChannelConfig()
        channel = _OutputChannelConfig()
        channel['name'] = name
        channel['channel'] = i
        channel['maxdata'] = 100
        channel['conversion-coefficients'] = (0,1)
        channel['conversion-origin'] = 0
        channel['inverse-conversion-coefficients'] = (0,1)
        channel['inverse-conversion-origin'] = 0
        piezo_config['inputs'].append(channel)
    return piezo_config



# Copyright (C) 2011-2012 W. Trevor King <wking@tremily.us>
#
# This file is part of pypiezo.
#
# pypiezo is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pypiezo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pypiezo.  If not, see <http://www.gnu.org/licenses/>.

"Piezo configuration"

import h5config.config as _config
import h5config.tools as _h5config_tools

import pycomedi.constant as _constant


class PackageConfig (_h5config_tools.PackageConfig):
    "Configure `pypiezo` module operation"
    settings = _h5config_tools.PackageConfig.settings + [
        _config.BooleanSetting(
            name='matplotlib',
            help='Plot piezo motion using `matplotlib`.',
            default=False),
        ]


class ChannelConfig (_config.Config):
    "Configure a single DAC/ADC channel"
    settings = [
        _config.Setting(
            name='name',
            help="Channel name (so the user will know what it's used for).",
            default=None),
        _config.Setting(
            name='device',
            help='Comedi device.',
            default='/dev/comedi0'),
        _config.IntegerSetting(
            name='subdevice',
            help='Comedi subdevice index.  -1 for automatic detection.',
            default=-1),
        _config.IntegerSetting(
            name='channel',
            help='Subdevice channel index.',
            default=0),
        _config.IntegerSetting(
            name='maxdata',
            help="Channel's maximum bit value."),
        _config.IntegerSetting(
            name='range',
            help="Channel's selected range index.",
            default=0),
        _config.ChoiceSetting(
            name='analog-reference',
            help="Channel's selected analog reference index.",
            choices=[(x.name, x) for x in _constant.AREF]),
        _config.FloatListSetting(
            name='conversion-coefficients',
            help=('Bit to physical unit conversion coefficients starting with '
                  'the constant coefficient.')),
        _config.FloatSetting(
            name='conversion-origin',
            help=('Origin (bit offset) of bit to physical polynomial '
                  'expansion.')),
        _config.FloatListSetting(
            name='inverse-conversion-coefficients',
            help=('Physical unit to bit conversion coefficients starting with '
                  'the constant coefficient.')),
        _config.FloatSetting(
            name='inverse-conversion-origin',
            help=('Origin (physical unit offset) of physical to bit '
                  'polynomial expansion.')),
        ]


class OutputChannelConfig (ChannelConfig):
    pass


class InputChannelConfig (ChannelConfig):
    pass


class AxisConfig (_config.Config):
    "Configure a single piezo axis"
    settings = [
        _config.FloatSetting(
            name='gain',
            help=(
                'Volts applied at piezo per volt output from the DAQ card '
                '(e.g. if your DAQ output is amplified before driving the '
                'piezo),')),
        _config.FloatSetting(
            name='sensitivity',
            help='Meters of piezo deflection per volt applied to the piezo.'),
        _config.FloatSetting(
            name='minimum',
            help='Set a lower limit on allowed output voltage',
            default=None),
        _config.FloatSetting(
            name='maximum',
            help='Set an upper limit on allowed output voltage',
            default=None),
        _config.ConfigSetting(
            name='channel',
            help='Configure the underlying DAC channel.',
            config_class=OutputChannelConfig,
            default=None),
        _config.ConfigSetting(
            name='monitor',
            help='Configure the underlying (optional) ADC monitoring channel.',
            config_class=InputChannelConfig,
            default=None),
        ]


class PiezoConfig (_config.Config):
    "Configure a piezo experiment"
    settings = [
        _config.Setting(
            name='name',
            help="Piezo name (so the user will know what it's used for).",
            default=None),
        _config.ConfigListSetting(
            name='axes',
            help='Configure the underlying axes.',
            config_class=AxisConfig,
            default=None),
        _config.ConfigListSetting(
            name='inputs',
            help='Configure the underlying (optional) ADC monitoring channels.',
            config_class=InputChannelConfig,
            default=None),
        ]


class WiggleConfig (_config.Config):
    "Configure an interference wiggle"
    settings = [
        _config.Setting(
            name='axis',
            help="Name of the axis to wiggle.",
            default='z'),
        _config.Setting(
            name='input',
            help="Name of the channel to watch.",
            default='deflection'),
        _config.FloatSetting(
            name='frequency',
            help="Frequency for a full wiggle cycle (hertz).",
            default=2),
        _config.IntegerSetting(
            name='samples',
            help='Number of samples in a full wiggle cycle.',
            default=1024),
        _config.IntegerSetting(
            name='offset',
            help='Center of the wiggle position (bits)',
            default=None),
        _config.IntegerSetting(
            name='amplitude',
            help='Amplitude of the wiggle sinusoid (bits)',
            default=None),
        _config.IntegerSetting(
            name='wavelength',
            help=('Instead of giving an explicit amplitude, you can specify '
                  'the laser wavelength in meters.  The amplitude will then '
                  'be set to contain about two interference cycles.'),
            default=None),
        ]
